@@ -46,6 +46,9 @@ typedef struct {
 	/* file to include */
 	gchar *include_file;
 
+	/* whether to own the included file */
+	gboolean own_file;
+
 	/* MiniDump file */
 	gchar *minidump_file;
 
